import subprocess

from i3configger.base import log


# todo use Adaephons i3 library
class I3:
    @classmethod
    def set_msg_type(cls, which):
        cls.refresh = {
            'restart': cls.restart_i3,
            'reload': cls.reload_i3,
        }.get(which, cls.nop)

    @classmethod
    def reload_i3(cls):
        cls._send_i3_msg('reload')

    @classmethod
    def restart_i3(cls):
        subprocess.call(['i3-msg', 'restart'])

    @classmethod
    def nop(cls):
        pass

    refresh = restart_i3

    @classmethod
    def _send_i3_msg(cls, msg):
        cmd = ['i3-msg', msg]
        try:
            output = subprocess.check_output(cmd).decode()
            if '"success":true' in output:
                return True
            return False
        except subprocess.CalledProcessError as e:
            if msg == 'restart' and e.returncode == 1:
                log.debug("[IGNORE] exit 1 is ok for restart")
                return True

    @classmethod
    def config_is_ok(cls, path):
        cmd = ['i3', '-C', str(path)]
        try:
            subprocess.check_output(cmd)
            return True
        except subprocess.CalledProcessError as e:
            Notify.send(f"error in {path}:\n{e.output}")
            return False


class Notify:
    @classmethod
    def set_notify_command(cls, offSwitch):
        if offSwitch:
            log.info("deactivate notification")
            cls.send = cls.nop

    @classmethod
    def send(cls, msg, urgency='low'):
        subprocess.check_call([
            'notify-send', '-a', 'i3configger', '-t', '1', '-u', urgency, msg])

    @staticmethod
    def nop(*args, **kwargs):
        pass


class StatusBar:
    @classmethod
    def refresh(cls):
        # TODO Hopefully this reloads the config?
        subprocess.check_call(['killall', '-SIGUSR1', 'i3status'])



import logging
import pprint
import re
import socket
import typing as t
from functools import total_ordering
from pathlib import Path

from i3configger import base, exc

log = logging.getLogger(__name__)

SPECIAL_SELECTORS = {
    "hostname": socket.gethostname()
}


@total_ordering
class Partial:
    CONTINUATION_RE = re.compile(r'\\\s*?\\s*?\n')
    COMMENT_MARK = '#'
    END_OF_LINE_COMMENT_MARK = ' # '

    def __init__(self, path: Path):
        self.path = path
        self.name = self.path.stem
        self.selectors = self.name.split('.')
        self.needsSelection = len(self.selectors) > 1
        self.key = self.selectors[0] if self.needsSelection else None
        self.value = self.selectors[1] if self.needsSelection else None

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.path.name)

    __str__ = __repr__

    def __lt__(self, other):
        return self.name < other.name

    @property
    def display(self) -> str:
        if not self.filtered:
            return ""
        return "### %s ###\n%s\n\n" % (self.path.name, self.filtered)

    @property
    def filtered(self) -> str:
        filtered = []
        for line in self._joined.splitlines():
            l = line.strip()
            if not l:
                continue
            if (not l.startswith(base.SET_MARK)
                    and not l.startswith(self.COMMENT_MARK)):
                filtered.append(line)
        return '\n'.join(filtered)

    @property
    def payload(self) -> str:
        """Strip empty lines, comment lines, and end of line comments."""
        prunes = []
        for line in self._joined.splitlines():
            l = line.strip()
            if not l:
                continue
            if l.startswith(self.COMMENT_MARK):
                continue
            line = line.rsplit(self.END_OF_LINE_COMMENT_MARK, maxsplit=1)[0]
            prunes.append(line)
        return '\n'.join(prunes)

    # FIXME I think this does not do anything
    @property
    def _joined(self) -> str:
        """Join line continuations.

        https://i3wm.org/docs/userguide.html#line_continuation"""
        return re.sub(self.CONTINUATION_RE, ' ', self._raw)

    @property
    def _raw(self) -> str:
        return self.path.read_text()


def find(prts: t.List[Partial], key: str, value: str= None) \
        -> t.Union[Partial, t.List[Partial]]:
    findings = []
    for prt in prts:
        if prt.key != key:
            continue
        if prt.value == value:
            return prt
        elif not value:
            findings.append(prt)
    return findings


def select(partials, selection, excludes=None) -> t.List[Partial]:
    def _select():
        selected.append(partial)
        if partial.needsSelection:
            del selection[partial.key]

    for key, value in SPECIAL_SELECTORS.items():
        if key not in selection:
            selection[key] = value
    selected = []
    for partial in partials:
        if partial.needsSelection:
            if excludes and partial.key in excludes:
                log.debug("[IGNORE] %s (in %s)", partial, excludes)
                continue
            if (selection and partial.key in selection and
                    partial.value == selection.get(partial.key)):
                _select()
        else:
            _select()
    log.debug("selected:\n%s", pprint.pformat(selected))
    if selection and not all(k in SPECIAL_SELECTORS for k in selection):
        raise exc.ConfigError(
            "selection processed incompletely: %s", selection)
    return selected


def create(partialsPath: Path) -> t.List[Partial]:
    assert partialsPath.is_dir(), partialsPath
    prts = [Partial(p) for p in partialsPath.glob('*%s' % base.SUFFIX)]
    if not prts:
        raise exc.PartialsError(f"no '*{base.SUFFIX}' at {partialsPath}")
    return sorted(prts)



import logging
import typing as t

from i3configger import base, exc, partials

log = logging.getLogger(__name__)


def create(prts: t.Iterable[partials.Partial]) -> dict:
    """Read variables from all partials into a dictionary and resolve them."""
    ctx = {}
    for context in [fetch(p.payload) for p in prts]:
        ctx.update(context)
    ctx = resolve(ctx)
    assert all(k.startswith(base.VAR_MARK) for k in ctx.keys()), ctx
    ctx = remove_variable_markers(ctx)
    assert not any(k.startswith(base.VAR_MARK) for k in ctx.keys()), ctx
    assert not any(v.startswith(base.VAR_MARK) for v in ctx.values()), ctx
    return ctx


def remove_variable_markers(ctx: dict) -> dict:
    cleaned = {}
    lvm = len(base.VAR_MARK)
    for key, value in ctx.items():
        key = key[lvm:] if key.startswith(base.VAR_MARK) else key
        cleaned[key] = value
    return cleaned


def enhance(ctx, otherContexts):
    eCtx = dict(ctx)
    for oc in otherContexts:
        if isinstance(oc, partials.Partial):
            oc = create([oc])
        eCtx.update(oc)
    return eCtx


def resolve(context: dict) -> dict:
    """If variables are set by a variable, replace them by their value."""
    if not any(v.startswith(base.VAR_MARK) for v in context.values()):
        log.debug("needs no resolution: %s", context)
        return context
    resolvedContext = {}
    seenKeys = []
    for key, value in context.items():
        if key in seenKeys:
            raise exc.DuplicateKey("%s -> %s would be overridden with %s",
                                   key, context[key], value)
        if value.startswith(base.VAR_MARK):
            try:
                resolvedContext[key] = context[value]
            except KeyError:
                log.exception("[IGNORED] %s, %s", key, value)
        else:
            resolvedContext[key] = value
    return resolvedContext


def fetch(content: str) -> dict:
    ctx = {}
    for line in content.splitlines():
        line = line.strip()
        if not line.startswith(base.SET_MARK):
            continue
        payload = line.split(maxsplit=1)[1]
        key, value = payload.split(maxsplit=1)
        ctx[key] = value
    return ctx



import logging
import pprint
import time
from pathlib import Path

from inotify import constants as ic
from inotify.adapters import Inotify

from i3configger import base, build, ipc

log = logging.getLogger(__name__)


class Watchman:
    BUILD_DELAY = 0.1
    """If an IDE does monkey business (e.g. Jetbrains "safe write")
    more than one change might be triggered for each change.

    Using a small delay to not trigger to often.
    """

    """Watch a config directories and build/refresh/notify on changes"""
    MASK = (
        ic.IN_CREATE | ic.IN_ATTRIB | ic.IN_DELETE |
        ic.IN_MODIFY | ic.IN_CLOSE_WRITE |
        ic.IN_MOVED_FROM | ic.IN_MOVED_TO |
        ic.IN_DELETE_SELF | ic.IN_MOVE_SELF)
    """Tell inotify to trigger on changes"""

    def __init__(self, configPath):
        self.partialsPath = str(configPath.parent).encode()
        self.configPath = configPath
        self.lastBuild = None
        self.lastFilePath = None
        self.errors = 0
        log.debug("initialized %s", self)

    def __str__(self):
        return "%s(%s)" % (
            self.__class__.__name__, pprint.pformat(self.__dict__))

    def watch(self):
        for event in self._get_events():
            self._process_event(event)

    def watch_guarded(self):
        for event in self._get_events():
            try:
                self._process_event(event)
            except:
                log.exception("I see dead calls ...")
                self.errors += 1
                if self.errors == 13:
                    log.critical("%s errors occurred, crashing", self.errors)
                    raise RuntimeError("%s: giving up" % self)

    def _get_events(self):
        inotify_watcher = Inotify()
        inotify_watcher.add_watch(self.partialsPath, mask=self.MASK)
        for event in inotify_watcher.event_gen():
            if not event:
                continue
            yield event

    def _process_event(self, event):
        header, typeNames, filePath = self._get_event_data(event)
        if self.needs_build(header, typeNames, filePath):
            log.info("%s triggered build", filePath)
            build.Builder(self.configPath).build()
            self.lastBuild = time.time()
            self.lastFilePath = filePath
            ipc.I3.refresh()
            ipc.StatusBar.refresh()
            ipc.Notify.send('Watchman: new config active')

    @staticmethod
    def _get_event_data(event):
        header, typeNames, watchPath, filename = event
        filePath = Path(watchPath.decode()) / filename.decode()
        log.debug("wd=%d|mask=%d|mask->names=%s|filePath=[%s]",
                  header.wd, header.mask, typeNames, filePath)
        return header, typeNames, filePath

    # noinspection PyUnusedLocal
    def needs_build(self, header, typeNames, filePath):
        if self.lastBuild and time.time() - self.lastBuild < self.BUILD_DELAY:
            log.debug("ignore %s changed too quick", filePath)
            return False
        if filePath.suffix not in [base.SUFFIX, '.json']:
            return False
        return True



import logging
import os
from pathlib import Path

from i3configger import config, exc

log = logging.getLogger(__name__)

MY_CONFIG_NAME = 'i3configger.json'
STATE_FILE_NAME = '.state.json'
MY_CONFIG_FOLDER = 'config.d'
MY_REL_CONFIG_PATH = MY_CONFIG_FOLDER + '/' + MY_CONFIG_NAME


class Paths:
    def __init__(self, configPath):
        path = Path(configPath)
        assert path.exists() and path.is_file(), path
        self.root = path.parent
        self.config = configPath
        self.state = self.root / STATE_FILE_NAME


def get_my_config_path(configPath=None):
    if not configPath:
        i3configPath = get_i3_config_path()
        configContainer = i3configPath / MY_CONFIG_FOLDER
        if not configContainer.exists():
            log.info("create new config folder at %s", configContainer)
            configContainer.mkdir()
        configPath = configContainer / MY_CONFIG_NAME
    elif configPath.is_dir():
            configPath /= MY_CONFIG_NAME
    if not configPath.exists():
        log.info("create default configuration at %s", configPath)
        config.freeze(configPath, config.I3_CONFIGGER_DEFAULTS)
    return configPath


def get_i3_config_path():
    """Use same search order like i3 (no system stuff though).

    see: https://github.com/i3/i3/blob/4.13/libi3/get_config_path.c#L31
    """
    candidates = [
        Path('~/.i3').expanduser(),
        Path(os.getenv("XDG_CONFIG_HOME", '~/.config/')).expanduser() / 'i3']
    for candidate in candidates:
        if candidate.exists():
            return candidate
    raise exc.ConfigError(
        f"can't find i3 config at the standard locations: {candidates}")



from pathlib import Path

import pytest
from pytest import fixture

from i3configger.build import Builder, MalformedAssignment


# noinspection PyShadowingNames
@fixture
def tmpdir(tmpdir):
    return Path(tmpdir)


@pytest.mark.parametrize(
    "line, exp",
    (
        ("set $var value", ("$var", "value")),
        (" set $var value", ("$var", "value")),
        (" set   $var #FFFFFF", ("$var", "#FFFFFF")),
        (" set   $var #FFFFFF # comment", ("$var", "#FFFFFF")),
        (" set   $var value # some comment", ("$var", "value")),
        ("set $var value other value  # comment ###",
         ("$var", "value other value")),
        ("set $var $otherVar  $otheVar", ("$var", "$otherVar  $otheVar")),
        (" set   $var #FFFFFF #malformed comment", MalformedAssignment),
        ("set spam eggs spam", MalformedAssignment),
        ("set", MalformedAssignment),
        ("set $var", MalformedAssignment),
    )
)
def test_get_assignment(line, exp):
    if not isinstance(exp, tuple):
        with pytest.raises(MalformedAssignment):
            Builder.get_assignment(line)
    else:
        assert Builder.get_assignment(line) == exp



import json
import logging
import pprint
from pathlib import Path

from i3configger import exc, partials, paths

log = logging.getLogger(__name__)

I3_CONFIGGER_DEFAULTS = {
  "main": {
    "target": "../config",
  },
  "bars": {
    "defaults": {
      "key": "i3status",
      "value": "full",
      "target": "..",
      "template": "tpl"
    },
    "targets": {}
  }
}


class I3configgerConfig:
    def __init__(self, configPath: Path, message: list=None):
        # TODO checks with helpful errors for screwed up configurations
        self.message = message
        p = paths.Paths(configPath)
        self.configPath = p.config
        self.partialsPath = p.root
        self.statePath = p.state
        self.payload = fetch(self.configPath)
        self.state = State.process(
            self.statePath, partials.create(self.partialsPath), self.message)
        targetPath = Path(self.payload["main"]["target"]).expanduser()
        if targetPath.is_absolute():
            self.mainTargetPath = targetPath.resolve()
        else:
            self.mainTargetPath = (self.partialsPath / targetPath).resolve()
        self.bars = self.populate_bar_defaults(self.payload.get("bars", {}))
        log.debug("initialized config  %s", self)

    def __str__(self):
        return "%s:\n%s" % (self.__class__.__name__,
                            pprint.pformat(vars(self)))

    def populate_bar_defaults(self, bars):
        """Create a resolved copy of the bar settings."""
        defaults = bars.get("defaults", {})
        resolvedBars = {}
        for name, bar in bars.items():
            if name == "defaults":
                continue
            newBar = dict(bar)
            resolvedBars[name] = newBar
            for defaultKey, defaultValue in defaults.items():
                if defaultKey not in newBar:
                    newBar[defaultKey] = defaultValue
        return resolvedBars


class State:
    """Process a message to alter the configuration to be build.

    Tuple contains name and number of expected additional arguments
    """
    I3STATUS = "i3status"
    """reserved key for status bar setting files"""
    SELECT_NEXT = ("select-next", 1)
    SELECT_PREVIOUS = ("select-previous", 1)
    SELECT = ("select", 2)
    SET = ("set", 2)
    _ALL = [SELECT_NEXT, SELECT_PREVIOUS, SELECT, SET]
    DEL = 'del'

    @classmethod
    def process(cls, statePath, prts, message):
        state = cls.fetch_state(statePath, prts)
        if not message:
            return state
        command, key, *rest = message
        value = rest[0] if rest else None
        log.debug(f"sending message {message} to {statePath}")
        if command == cls.SET[0]:
            if value.lower() == cls.DEL:
                del state["set"][key]
            else:
                state["set"][key] = value
        else:
            if command in [cls.SELECT_NEXT[0], cls.SELECT_PREVIOUS[0]]:
                candidates = partials.find(prts, key)
                if not candidates:
                    raise exc.MessageError(
                        f"No candidates for {message} in {prts}")
                if command == cls.SELECT_PREVIOUS:
                    candidates = reversed(candidates)
                current = state["select"].get(key) or candidates[0].key
                for idx, candidate in enumerate(candidates):
                    if candidate.value == current:
                        try:
                            new = candidates[idx + 1]
                        except IndexError:
                            new = candidates[0]
                        log.info("select %s.%s", key, new)
                        state["select"][key] = new.value
                        break
            elif command == cls.SELECT:
                candidate = partials.find(prts, key, value)
                if not candidate:
                    raise exc.MessageError(f"No candidate for {message}")
                if value.lower == cls.DEL:
                    del state["select"][key]
                else:
                    state["select"][key] = candidate.value
        freeze(statePath, state)
        return state

    @classmethod
    def fetch_state(cls, statePath, prts):
        if not statePath.exists():
            return cls.populate_initial_state(statePath, prts)
        return fetch(statePath)

    @classmethod
    def populate_initial_state(cls, statePath, prts):
        """fetch one of each selectable partials to have a sane state"""
        selects = {}
        for prt in prts:
            if prt.needsSelection and not prt.key == cls.I3STATUS:
                selects[prt.key] = prt.value
        state = dict(select=selects, set={})
        freeze(statePath, state)
        return state

    @classmethod
    def get_spec(cls, message):
        for c in cls._ALL:
            if c[0] == message:
                return c[1]
        raise exc.MessageError(f"unknown message: {message}. "
                               f"I only now: {[m[0] for m in cls._ALL]}")


def fetch(path):
    if not path.exists():
        raise exc.ConfigError(f"No config found at {path}")
    with path.open() as f:
        log.info("read config from %s", path)
        payload = json.load(f)
        log.debug("use:\n%s", pprint.pformat(payload))
        return payload


def freeze(path, obj):
    with open(path, 'w') as f:
        json.dump(obj, f, sort_keys=True, indent=4)
    log.debug("froze %s to %s", pprint.pformat(obj), path)



from pathlib import Path

import pytest

from i3configger import build, paths

HERE = Path(__file__).parent
EXAMPLES = HERE.parent / 'examples'


@pytest.mark.parametrize(
    "container", (
        "default", "selectors"
    ))
def test_no_config(container, monkeypatch):
    """Given empty sources directory a new config is created from defaults"""
    pytest.skip("WIP")
    container = EXAMPLES / container
    monkeypatch.setattr(paths, 'get_i3_config_path', lambda: container)
    configPath = paths.get_my_config_path()
    assert configPath.exists() and configPath.is_file()

    builder = build.Builder(configPath)
    builder.build()
    assert 0



import pytest

from i3configger import cli, exc


@pytest.mark.parametrize(
    "args,exp",
    (
        ([], True),
        ([''], False),
        (['dunno'], False),
        (['select-next'], False),
        (['select-previous'], False),
        (['set'], False),
        (['select'], False),
        # TODO do I want special stuff to be overridable?
        (['select', 'hostname', 'ob1'], True),
        (['set', 'someVar', 'someValue'], True),
        (['set', 'someVar'], False),
        (['select-next', 'scheme'], True),
        (['select-next', 'scheme', 'extra'], False),
        (['select-previous', 'scheme'], True),
        (['select-previous', 'scheme', 'extra'], False),
    )
)
def test_check_sanity(args, exp):
    if exp:
        cli.check_sanity(args)
    else:
        with pytest.raises(exc.I3configgerException):
            cli.check_sanity(args)



from setuptools import find_packages, setup

try:
    import pypandoc
    longDescription = pypandoc.convert('README.md', 'rst')
except(OSError, IOError, ImportError):
    longDescription = open('README.md').read()


kwargs = dict(
    name='i3configger',
    author='Oliver Bestwalter',
    url='https://github.com/obestwalter/i3configger',
    description="i3 config generation tool",
    long_description=longDescription,
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    entry_points={'console_scripts': ['i3configger = i3configger.main:main']},
    install_requires=['inotify', 'psutil', 'python-daemon'],
    packages=find_packages(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: POSIX :: Linux',
        'License :: OSI Approved :: MIT License',
        'Environment :: Console',
        'Topic :: Utilities',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
    ],
)


if __name__ == '__main__':
    setup(**kwargs)



"""This is just a sketch ... """
raise NotImplementedError()

BINDCODE = 'bindcode'
BINDSYM = 'bindsym'


# set after build
class Bindings:
    """
    bindsym | bindcode
    [--release] [<Group>+][<Modifiers>+]<keysym> command

    [--release] [--border] [--whole-window] [<Modifiers>+]button<n> command
    """
    def __init__(self, lines):
        self.lines = lines

    def translate_bindings(self):
        """translate bindcode to bindsym assignments

        this need to be done the moment the information is asked because it
        depends on the currently acitve layout.
        """

    def write_bindings_info(self):
        """Write info in some format that can be nicely displayed"""



import argparse
import logging
from pathlib import Path

from i3configger import __version__, base, exc, config

log = logging.getLogger(__name__)


def process_command_line():
    parser = argparse.ArgumentParser(
        'i3configger',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    args = _parse_args(parser)
    base.configure_logging(verbosity=args.v, logPath=args.log)
    check_sanity(args.message)
    args.config = Path(args.config).expanduser() if args.config else None
    if args.message and any([args.daemon, args.kill, args.watch]):
        parser.error(
            "message and daemon/watch can't be used together. "
            "Start the watcher process first and then you can send messages"
            "in following calls.")
    if args.load_config and any([args.daemon, args.kill, args.watch]):
        parser.error(
            "Loading a configuration or state and daemon/watch can't be "
            "used together. Start the watcher process first and then load"
            "in following calls.")
    return args


def _parse_args(p):
    """Command line commands - all optional with [reasonable] defaults"""
    p.add_argument('--version', action='version',
                   version="%s %s" % (p.prog, __version__))
    p.add_argument('-v', action="count", help="raise verbosity", default=0)
    g = p.add_mutually_exclusive_group()
    g.add_argument('--watch', action="store_true",
                   help="watch and build in foreground", default=False)
    g.add_argument('--daemon', action="store_true",
                   help="watch and build as daemon", default=False)
    g.add_argument('--kill', action="store_true", default=False,
                   help="exorcise daemon if running")
    p.add_argument('--i3-refresh-msg', action="store", default='reload',
                   choices=['restart', 'reload', 'nop'],
                   help="i3-msg to send after build")
    p.add_argument('--no-notify', action="store_true", default=False,
                   help="deactivate notification via notify-send")
    p.add_argument('--log', action="store", default=None,
                   help="i3configgerPath to where log should be stored")
    p.add_argument('--load-config', action="store", default=None,
                   help="load a config and build new")
    p.add_argument('-c', '--config', action="store",
                   default=None, help="i3configgerPath to config file")
    p.add_argument("message", help="message to send to i3configger", nargs="*")
    return p.parse_args()


def check_sanity(message):
    """Extract i3configger commands."""
    if not message:
        return
    log.debug("processing %s", message)
    command, *rest = message
    spec = config.State.get_spec(command)
    if len(rest) != spec:
        raise exc.I3configgerException(
            f"message '{command}' needs {spec} args - got: {rest}")
    log.debug(f"{message} seem to be a legit message")



import logging
import os
import sys
import tempfile
from pathlib import Path

from i3configger import exc

log = logging.getLogger(__name__)
DEBUG = os.getenv('DEBUG', 1)  # FIXME deactivate before next release
VAR_MARK = '$'
SET_MARK = 'set'
SUFFIX = '.conf'
SETTINGS_MARK = VAR_MARK + SET_MARK + '_'


def configure_logging(verbosity: int, logPath: str, isDaemon=False):
    rootLogger = logging.getLogger()
    _log = str(Path(tempfile.gettempdir()) / 'i3configger.log')
    logPath = Path(logPath).expanduser() if logPath else _log
    if DEBUG:
        print('logging to %s' % logPath)
        level = 'DEBUG'
    else:
        level = logging.getLevelName(
            {0: 'ERROR', 1: 'WARNING', 2: 'INFO'}.get(verbosity, 'DEBUG'))
    fmt = ('%(asctime)s %(name)s:%(funcName)s:%(lineno)s '
           '%(levelname)s: %(message)s')
    if isDaemon:
        logging.basicConfig(filename=logPath, format=fmt, level=level)
    else:
        logging.basicConfig(format=fmt, level=level)
        fileHandler = logging.FileHandler(logPath)
        fileHandler.setFormatter(logging.Formatter(fmt))
        fileHandler.setLevel(level)
        rootLogger.addHandler(fileHandler)


def i3configger_excepthook(type_, value, traceback):
    if DEBUG or not isinstance(value, exc.I3configgerException):
        _REAL_EXCEPTHOOK(type_, value, traceback)
    else:
        sys.exit("%s: %s" % (value.__class__.__name__, value))


_REAL_EXCEPTHOOK = sys.excepthook
sys.excepthook = i3configger_excepthook



import logging
import sys
from pathlib import Path

from i3configger import (
    build, cli, config, daemonize, ipc, partials, paths, watch)

log = logging.getLogger(__name__)


def main():
    args = cli.process_command_line()
    if args.kill:
        daemonize.exorcise()
        return 0
    configPath = paths.get_my_config_path(args.config)
    if args.message:
        p = paths.Paths(configPath)
        config.State.process(p.state, partials.create(p.root), args.message)
    if daemonize.get_other_i3configgers():
        if not args.message:
            sys.exit("i3configger already running. "
                     "Did you want to send a message?")
        # process and write the state and let watcher do the build
        log.info("watcher process is running - processing message")
        return 0
    ipc.I3.set_msg_type(args.i3_refresh_msg)
    log.info("set i3 refresh method to %s", ipc.I3.refresh)
    ipc.Notify.set_notify_command(args.no_notify)
    if args.daemon:
        daemonize.daemonize(args.v, args.log, configPath)
        return 0
    if args.watch:
        try:
            watch.Watchman(configPath).watch()
        except KeyboardInterrupt:
            sys.exit("bye")
    if args.load_config:
        p = paths.Paths(configPath)
        state = config.State.fetch_state(
            Path(args.load).expanduser(), partials.create(p.root))
        config.freeze(configPath, state)
    else:
        build.Builder(configPath).build()
        ipc.I3.refresh()
        ipc.StatusBar.refresh()
        ipc.Notify.send('new config active')



import logging
import os
import pprint
import time
from pathlib import Path
from string import Template

from i3configger import base, config, context, exc, ipc, partials

log = logging.getLogger(__name__)


class Builder:
    def __init__(self, configPath):
        self.cnf = config.I3configgerConfig(configPath)
        log.info("initialized %s", self)

    def __str__(self):
        return "%s\n%s" % (self.__class__.__name__, pprint.pformat(vars(self)))

    def build(self):
        content = self._build()
        self.persist_main(content, self.cnf.mainTargetPath)

    def _build(self):
        prts = partials.create(self.cnf.partialsPath)
        excludes = {b["key"] for b in self.cnf.bars.values()}
        selected = partials.select(prts, self.cnf.state["select"], excludes)
        if not selected:
            raise exc.I3configgerException(
                "no content for %s, %s, %s", prts, self.cnf)
        ctx = context.create(selected)
        rawContent = self.make_header()
        rawContent += '\n'.join(prt.display for prt in selected)
        resolvedContent = self.substitute(rawContent, ctx)
        if not self.cnf.bars:
            return resolvedContent
        barContent = self.get_bar_content(prts, ctx, self.cnf.state)
        return "%s\n%s" % (resolvedContent, barContent)

    def persist_main(self, content, path):
        container = path.parent
        targetName = path.name
        backupPath = container / (targetName + '.bak')
        os.rename(self.cnf.mainTargetPath, backupPath)
        try:
            self.cnf.mainTargetPath.write_text(content)
            if not ipc.I3.config_is_ok(self.cnf.mainTargetPath):
                brokenPath = container / (targetName + '.broken')
                os.rename(self.cnf.mainTargetPath, brokenPath)
                raise exc.BuildError(f"{brokenPath} is broken")
        except:
            os.rename(backupPath, self.cnf.mainTargetPath)
            raise

    def get_bar_content(self, prts, ctx, state):
        bars = []
        alreadyWritten = []
        for barName, barCnf in self.cnf.bars.items():
            barCnf["id"] = barName
            selectKey = barCnf["key"]
            selectValue = barCnf["value"]
            prt = partials.find(prts, selectKey, selectValue)
            assert isinstance(prt, partials.Partial), prt
            tpl = partials.find(prts, selectKey, barCnf["template"])
            assert isinstance(tpl, partials.Partial), tpl
            container = Path(barCnf["target"])
            if not container.is_absolute():
                container = (self.cnf.partialsPath / container).resolve()
                barCnf["target"] = str(container)
            eCtx = context.enhance(ctx, [barCnf, prt, state["set"]])
            bars.append(self.substitute(tpl.display, eCtx))
            if prt.name not in alreadyWritten:
                content = self.substitute(prt.payload, eCtx)
                path = container / f"{prt.name}{base.SUFFIX}"
                path.write_text(content)
                alreadyWritten.append(path)
        return '\n'.join(bars)

    @classmethod
    def substitute(cls, content, ctx):
        """Substitute all variables with their values.

        Works out of the box, because '$' is the standard substitution
        marker for string.Template
        """
        return Template(content).safe_substitute(ctx)

    def make_header(self):
        msg = f'# Generated from {self.cnf.configPath} ({time.asctime()}) #'
        sep = "#" * len(msg)
        return "%s\n%s\n%s" % (sep, msg, sep)



from pathlib import Path

from i3configger.config import I3configgerConfig

DATA = Path(__file__).parent / 'data'


def test_no_file():
    I3configgerConfig.NAME = 'i-do-not-exist'
    i3s = I3configgerConfig(DATA)
    assert not i3s
    assert not hasattr(i3s, 'bars')
    assert not hasattr(i3s, 'defaults')


def test_empty_file():
    I3configgerConfig.NAME = 'i3status-empty.json'
    i3s = I3configgerConfig(DATA)
    assert not i3s
    assert not hasattr(i3s, 'bars')
    assert not hasattr(i3s, 'defaults')


def test_only_bars():
    I3configgerConfig.NAME = 'i3status-only-bars.json'
    i3s = I3configgerConfig(DATA)
    assert i3s
    assert i3s.bars == {"laptop": {"output": "DP-3", "position": "top"}}
    assert i3s.marker == i3s.DEFAULT_SETTINGS[i3s.MARKER_KEY]
    assert i3s.template == i3s.DEFAULT_SETTINGS[i3s.TEMPLATE]
    assert i3s.target == i3s.DEFAULT_SETTINGS[i3s.TARGET]
    assert i3s.defaults == {}


def test_full():
    I3configgerConfig.NAME = 'i3status-full.json'
    i3s = I3configgerConfig(DATA)
    assert i3s
    assert i3s.bars == {
        'first bar': {
            'font': 'should override font default',
            'output': 'first output',
            'source': 'some default source'},
        'second bar': {
            'font': 'some default font',
            'output': 'second output',
            'source': 'should override source default'}}
    assert i3s.marker == "not marker default"
    assert i3s.template == "not template default"
    assert i3s.target == i3s.DEFAULT_SETTINGS[i3s.TARGET]
    assert i3s.defaults == {"font": "some default font",
                            "source": "some default source"}
